{==============================================================================}
{     Label
{==============================================================================}

unit CLabels;

interface

uses Windows, SysUtils, Messages, ShellAPI, Classes, Graphics, Controls, Forms,
   StdCtrls, ExtCtrls;

type

TOHUrlLabel = class(TLabel)
   private
      FURL        : string;
      FLinkColor  : TColor;
      FALinkColor : TColor;
      FHotColor   : TColor;
      FVLinkColor : TColor;

      FVisited    : Boolean;
      FDown       : Boolean;
      FOver       : Boolean;

      procedure SetLinkColor (Value : TColor);
      procedure SetVLinkColor (Value : TColor);
      procedure CMFontChanged (var m : TMessage); message CM_FONTCHANGED;
      procedure CMMouseEnter (var m : TMessage); message CM_MOUSEENTER;
      procedure CMMouseLeave (var m : TMessage); message CM_MOUSELEAVE;

   protected
      procedure MouseDown (Button : TMouseButton; Shift : TShiftState;
         X, Y : Integer); override;
      procedure MouseUp (Button : TMouseButton; Shift : TShiftState;
         X, Y : Integer); override;

   public
      constructor Create (AOwner : TComponent); override;

   published
      property Cursor      default crHandPoint;
      property URL         : string read FURL        write FURL;
      property LinkColor   : TColor read FLinkColor  write SetLinkColor  default clBlue;
      property ALinkColor  : TColor read FALinkColor write FALinkColor   default clRed;
      property VLinkColor  : TColor read FVLinkColor write SetVLinkColor default clNavy;
      property HotColor    : TColor read FHotColor   write FHotColor     default clRed;

   end;

TOHHotLabel = class(TLabel)
   private
      FHotColor   : TColor;

      FLinkColor  : TColor;

      procedure CMMouseEnter (var m : TMessage); message CM_MOUSEENTER;
      procedure CMMouseLeave (var m : TMessage); message CM_MOUSELEAVE;

   protected

   public
      constructor Create (AOwner : TComponent); override;

   published
      property Cursor      default crHandPoint;
      property HotColor    : TColor read FHotColor write FHotColor  default clMaroon;

   end;


TOH_SymbLabel = class(TCustomLabel)
   private
      FSmallFont        : TFont;
      FStrongFont       : TFont;
      FSymbolFont       : TFont;

      procedure Changed (Sender : TObject);
      procedure SetSmallFont (Value : TFont);
      procedure SetStrongFont (Value : TFont);
      procedure SetSymbolFont (Value : TFont);

   protected
      procedure SwitchFont (var AText : string; var VAlign : TTextLayout);
      function  SymbolTextSize (const AText : string; const ARect : TRect) : TSize;
      function  SymbolTextRect (const AText : string; const ARect : TRect; AAlign : TAlignment) : TRect;
      procedure SymbolTextDraw (const AText : string; const ARect : TRect; AAlign : TAlignment);
      procedure DrawLabelText (var ARect : TRect; AAlign : TAlignment; const AText : string);
      procedure Paint; override;

   public
      constructor Create (AOwner : TComponent); override;
      destructor Destroy; override;

      property SmallFont      : TFont  read FSmallFont    write SetSmallFont;
      property StrongFont     : TFont  read FStrongFont   write SetStrongFont;
      property SymbolFont     : TFont  read FSymbolFont   write SetSymbolFont;

   end;

TOHSymbLabel = class(TOH_SymbLabel)
   published
      property Align;
      property Alignment;
      property Anchors;
      property AutoSize;
      property BiDiMode;
      property Caption;
      property Color;
      property Constraints;
      property DragCursor;
      property DragKind;
      property DragMode;
      property Enabled;
      property FocusControl;
      property Font;
      property ParentBiDiMode;
      property ParentColor;
      property ParentFont;
      property ParentShowHint;
      property PopupMenu;
      property ShowAccelChar;
      property ShowHint;
      property SmallFont;
      property StrongFont;
      property SymbolFont;
      property Transparent;
      property Layout;
      property Visible;
//      property WordWrap;
      property OnClick;
      property OnContextPopup;
      property OnDblClick;
      property OnDragDrop;
      property OnDragOver;
      property OnEndDock;
      property OnEndDrag;
      property OnMouseDown;
      property OnMouseMove;
      property OnMouseUp;
      property OnStartDock;
      property OnStartDrag;

   end;

TOHReadoutLabel = class(TOH_SymbLabel)
   private
      FValueText        : string;
      FValueWidth       : Integer;
      FUnitsText        : string;
      FUnitsWidth       : Integer;

      procedure SetValueText (const Value : string);
      procedure SetValueWidth (Value : Integer);
      procedure SetUnitsText (const Value : string);
      procedure SetUnitsWidth (Value : Integer);

   protected
      procedure Paint; override;

   public
      constructor Create (AOwner : TComponent); override;

   published
      property Align;
//      property Alignment;
      property Anchors;
//      property AutoSize;
      property BiDiMode;
      property Caption;
      property Color;//       default clWindow;
      property Constraints;
      property DragCursor;
      property DragKind;
      property DragMode;
      property Enabled;
//      property FocusControl;
      property Font;
      property ParentBiDiMode;
      property ParentColor;
      property ParentFont;
      property ParentShowHint;
      property PopupMenu;
//      property ShowAccelChar;
      property ShowHint;
      property SmallFont;
      property StrongFont;
      property SymbolFont;
//      property Transparent;
//      property Layout;
      property Visible;
      property ValueText      : string    read FValueText   write SetValueText;
      property ValueWidth     : Integer   read FValueWidth  write SetValueWidth;
      property UnitsText      : string    read FUnitsText   write SetUnitsText;
      property UnitsWidth     : Integer   read FUnitsWidth  write SetUnitsWidth;
//      property WordWrap;
      property OnClick;
      property OnContextPopup;
      property OnDblClick;
      property OnDragDrop;
      property OnDragOver;
      property OnEndDock;
      property OnEndDrag;
      property OnMouseDown;
      property OnMouseMove;
      property OnMouseUp;
      property OnStartDock;
      property OnStartDrag;

   end;


const
   DefaultItemHeight = 20;

type

TOHListLabel = class;

TOHListColumn = class(TCollectionItem)
   private
      FAlignment     : TAlignment;
      FColor         : TColor;
      FFont          : TFont;
      FLined         : Boolean;
      FParentColor   : Boolean;
      FParentFont    : Boolean;
      FWidth         : Integer;

      procedure FontChanged (Sender : TObject);
      function  GetDisplayFont : TFont;
      function  IsColorStored : Boolean;
      function  IsFontStored : Boolean;
      procedure SetAlignment (Value : TAlignment);
      procedure SetColor (Value : TColor);
      procedure SetFont (Value : TFont);
      procedure SetLined (Value : Boolean);
      procedure SetParentColor (Value : Boolean);
      procedure SetParentFont (Value : Boolean);
      procedure SetWidth (Value : Integer);

   protected
      function  GetList : TOHListLabel;

   public
      constructor Create (Collection : TCollection); override;
      destructor Destroy; override;
      procedure Assign (Source : TPersistent); override;
      function  DefaultColor : TColor;
      function  DefaultFont : TFont;
      procedure RestoreDefaults; virtual;

      property  List       : TOHListLabel read GetList;

   published
      property  Alignment  : TAlignment   read FAlignment   write SetAlignment   default taLeftJustify;
      property  Color      : TColor       read FColor {GetColor} write SetColor  stored IsColorStored    default clNone;
      property  Font       : TFont        read FFont        write SetFont        stored IsFontStored;
      property  Lined      : Boolean      read FLined       write SetLined       default False;
      property  ParentColor: Boolean      read FParentColor write SetParentColor default True;
      property  ParentFont : Boolean      read FParentFont  write SetParentFont  default True;
      property  Width      : Integer      read FWidth       write SetWidth;

   end;


TOHListColumnClass = class of TOHListColumn;

TOHListColumns = class(TCollection)
   private
      FList       : TOHListLabel;

      function GetColumn (Index : Integer) : TOHListColumn;
      procedure SetColumn (Index : Integer; Value : TOHListColumn);

   protected
      function GetOwner : TPersistent; override; {D3}
      procedure Update (Item : TCollectionItem); override;

   public
      constructor Create (List : TOHListLabel; ColumnClass: TOHListColumnClass);
      function  Add : TOHListColumn;
      procedure LoadFromFile (const Filename : string);
      procedure LoadFromStream (S : TStream);
      procedure RestoreDefaults;
      procedure SaveToFile (const Filename : string);
      procedure SaveToStream (S : TStream);

      property List : TOHListLabel read FList;
      property Items[Index : Integer]: TOHListColumn read GetColumn write SetColumn; default;

   end;


TOHListLabel = class(TCustomLabel)
   private
      FBevel         : TPanelBevel;
      FBevelWidth    : Integer;
      FBorderCols    : Integer;
      FBorderRows    : Integer;
      FBorderStyle   : TBorderStyle;
//      FBorderWidth   : Integer;
      FColumns       : TOHListColumns;
      FItemHeight    : Integer;
      FItems         : TStringList;

      function  WidthToColumn (ACol : Integer) : Integer;
      function  HeightToRow (ARow : Integer) : Integer;
      procedure ItemsChanged (Sender : TObject);
      procedure SetBevel (Value : TPanelBevel);
      procedure SetBevelWidth (Value : Integer);
      procedure SetBorderCols (Value : Integer);
      procedure SetBorderRows (Value : Integer);
      procedure SetBorderStyle (Value : TBorderStyle);
//      procedure SetBorderWidth (Value : Integer);
      function  GetColCount : Integer;
      procedure SetColCount (Value : Integer);
      procedure SetColumns (Value : TOHListColumns);
      procedure SetItemHeight (Value : Integer);
      procedure SetItems (Value : TStringList);
      function  GetRowCount : Integer;

   protected
      function  CreateColumns : TOHListColumns; dynamic;
      procedure AdjustBounds; override;
//      procedure Loaded; override;
      procedure Paint; override;

   public
      constructor Create (AOwner : TComponent); override;
      destructor Destroy; override;

    procedure Clear;

      property RowCount       : Integer         read GetRowCount;

   published
      property Align;
      property Alignment;
      property AutoSize;
      property Bevel          : TPanelBevel     read FBevel      write SetBevel        default bvNone;
      property BevelWidth     : Integer         read FBevelWidth write SetBevelWidth   default 1;
      property BorderCols     : Integer         read FBorderCols write SetBorderCols   default 0;
      property BorderRows     : Integer         read FBorderRows write SetBorderRows   default 0;
      property BorderStyle    : TBorderStyle    read FBorderStyle write SetBorderStyle default bsNone;
//      property BorderWidth    : Integer         read FBorderWidth write SetBorderWidth;
      property Caption;
      property ColCount       : Integer         read GetColCount write SetColCount;
      property Columns        : TOHListColumns  read FColumns    write SetColumns;
      property Color;
      property DragCursor;
      property DragMode;
      property Enabled;
      property FocusControl;
      property Font;
      property ItemHeight     : Integer         read FItemHeight write SetItemHeight default 0;
      property Items          : TStringList     read FItems      write SetItems;
      property Height;
      property Layout;
      property Left;
      property ParentColor;
      property ParentFont;
      property ParentShowHint;
      property PopupMenu;
      property ShowAccelChar;
      property ShowHint;
      property Visible;
      property Top;
      property Transparent;
      property Width;
      property WordWrap;
      property OnClick;
      property OnDblClick;
      property OnDragDrop;
      property OnDragOver;
      property OnEndDrag;
      property OnMouseDown;
      property OnMouseMove;
      property OnMouseUp;

   end;

procedure Register;


{==============================================================================}
{==============================================================================}


implementation

{------------------------------------------------------------------------------}

function Uti_SubStr (const d : string; var s : string) : string;
{ return a sub string
'
'     d$    - delimiting char
'
'     s$    - main string
'
'           - return sub-string
}
var i : Integer;
begin

   i := Pos(d, s);
   if i = 0 then begin
      Result := s;
      s := '';
      end
   else begin
      Result := Copy(s, 1, i - 1);
      s := Copy(s, i + Length(d), Length(s));
      end;

end;

{------------------------------------------------------------------------------}

function Gfx_LimitColourByte (b : Integer) : Integer;
begin

        if b > 255 then Result := 255
   else if b < 0   then Result := 0
                   else Result := b;

end;

{------------------------------------------------------------------------------}

function Gfx_BrightenColour (c : TColor; d : Integer) : TColor;
begin

	c := ColorToRGB(c);

	Result := RGB(Gfx_LimitColourByte(GetRValue(c) + d),
                 Gfx_LimitColourByte(GetGValue(c) + d),
                 Gfx_LimitColourByte(GetBValue(c) + d));

end;

{------------------------------------------------------------------------------}

procedure Gfx_BevelColours (d : Boolean; c : TColor; var c1, c2 : TColor);
begin

   if d then begin
      c1 := Gfx_BrightenColour(c, -64);
      c2 := Gfx_BrightenColour(c,  64);
      end
   else begin
      c1 := Gfx_BrightenColour(c,  64);
      c2 := Gfx_BrightenColour(c, -64);
      end;

end;

{==============================================================================}

constructor TOHUrlLabel.Create(AOwner : TComponent);
begin

   inherited Create(AOwner);

   Font.Color   := clBlue;
   Font.Style   := [fsUnderline];
   FLinkColor   := clBlue;
   FALinkColor  := clRed;
   FVLinkColor  := clNavy;
   FHotColor    := clRed;
   Cursor       := crHandPoint;

end;

{------------------------------------------------------------------------------}

procedure TOHUrlLabel.CMFontChanged (var m : TMessage);
begin

   inherited;

   if not FDown and not FOver then begin
      if not FVisited then
         FLinkColor := Font.Color
      else
         FVLinkColor := Font.Color;
      end;

end;

{------------------------------------------------------------------------------}

procedure TOHUrlLabel.MouseDown (Button : TMouseButton; Shift : TShiftState;
   X, Y : Integer);
begin
   inherited MouseDown(Button, Shift, X, Y);

   FDown := True;
   Font.Color := FALinkColor;
end;

{------------------------------------------------------------------------------}


procedure TOHUrlLabel.MouseUp (Button : TMouseButton; Shift : TShiftState;
   X, Y : Integer);

var u : string;
begin
   inherited MouseUp(Button, Shift, X, Y);

   if (X >= 0) and (Y >= 0) and (Y <= Height) and (X <= Width) then begin
      if FDown then begin
         FVisited := True;
         Font.Color := FVLinkColor;
         u := FURL; if u = '' then u := Caption;
         ShellExecute(ValidParentForm(Self).Handle, 'open',
            PChar(u), nil, nil, SW_SHOW);
         end;
      end
   else begin
      if FVisited then
         Font.Color := FVLinkColor
      else
         Font.Color := FLinkColor;
      end;

   FDown := False;

end;

{------------------------------------------------------------------------------}

procedure TOHUrlLabel.CMMouseEnter (var m : TMessage);
begin
   inherited;

   FOver := True;
   if FDown then Font.Color := FALinkColor
            else Font.Color := FHotColor;
end;

{------------------------------------------------------------------------------}

procedure TOHUrlLabel.CMMouseLeave (var m : TMessage);
begin
   inherited;

   if FVisited then
      Font.Color := FVLinkColor
   else
      Font.Color := FLinkColor;
   FOver := False;
end;

{------------------------------------------------------------------------------}

procedure TOHUrlLabel.SetLinkColor (Value : TColor);
begin

   if FLinkColor <> Value then begin
      if not FVisited then
         Font.Color := Value
      else
         FLinkColor := Value;
      end;

end;

{------------------------------------------------------------------------------}

procedure TOHUrlLabel.SetVLinkColor (Value : TColor);
begin

   if FVLinkColor <> Value then begin
      if not FVisited then
         FVLinkColor := Value
      else
         Font.Color := Value;
      end;

end;

{==============================================================================}

constructor TOHHotLabel.Create(AOwner : TComponent);
begin

   inherited Create(AOwner);

   Font.Color   := clNavy;
   FHotColor    := clMaroon;
   Cursor       := crHandPoint;

end;

{------------------------------------------------------------------------------}

procedure TOHHotLabel.CMMouseEnter (var m : TMessage);
begin
   inherited;

   FLinkColor := Font.Color;
   Font.Color := FHotColor;
end;

{------------------------------------------------------------------------------}

procedure TOHHotLabel.CMMouseLeave (var m : TMessage);
begin
   inherited;

   Font.Color := FLinkColor;
end;


{==============================================================================}
{     Symbol Controls
{==============================================================================}

constructor TOH_SymbLabel.Create (AOwner : TComponent);
begin

   inherited Create(AOwner);

   FSmallFont       := TFont.Create;
   FSmallFont.OnChange := Changed;
   FSmallFont.Name  := 'Small Fonts';
   FSmallFont.Size  := 6;

   FStrongFont       := TFont.Create;
   FStrongFont.OnChange := Changed;
   FStrongFont.Color := clRed;

   FSymbolFont       := TFont.Create;
   FSymbolFont.OnChange := Changed;
   FSymbolFont.Name  := 'Symbol';

end;

{------------------------------------------------------------------------------}

destructor TOH_SymbLabel.Destroy;
begin
   FSmallFont.Free;
   FSymbolFont.Free;

   inherited Destroy;
end;

{------------------------------------------------------------------------------}

procedure TOH_SymbLabel.SwitchFont (var AText : string; var VAlign : TTextLayout);
begin

   VAlign := tlTop;

      case AText[1] of
   '~' : begin
         Canvas.Font := FSymbolFont;
         Delete(AText, 1, 1);
         end;
   '!' : begin
         Canvas.Font := FStrongFont;
         Delete(AText, 1, 1);
         end;
   '<' : begin
         Canvas.Font := FSmallFont;
         Delete(AText, 1, 1);
         VAlign := tlBottom;
         end;
   '>' : begin
         Canvas.Font := FSmallFont;
         Delete(AText, 1, 1);
         end;
   else
         Canvas.Font := Font;
      end;

end;

{------------------------------------------------------------------------------}

function TOH_SymbLabel.SymbolTextSize (const AText : string; const ARect : TRect) : TSize;

var a, t : string;
    r : TRect;
    h : Integer;
    v : TTextLayout;
begin

   with Result do begin cx := 0;  cy := 0 end;
   t := AText;
   repeat
      a := Uti_SubStr('$', t);
      if a <> '' then begin
         SwitchFont(a, v);
         r := ARect;
         DrawText(Canvas.Handle, PChar(a), Length(a), r, DT_EXPANDTABS or DT_CALCRECT);
         Inc(Result.cx, r.Right - r.Left);
         h := (r.Bottom - r.Top) * 4 div 3;
         if h > Result.cy then Result.cy := h;
         end;
      until t = '';

end;

{------------------------------------------------------------------------------}

function TOH_SymbLabel.SymbolTextRect (const AText : string; const ARect : TRect; AAlign : TAlignment) : TRect;

var z : TSize;
begin

   Result := ARect;

   z := SymbolTextSize(AText, ARect);
      case AAlign of
   taLeftJustify :
         Result.Right := Result.Left  + z.cx;
   taRightJustify :
         Result.Left  := Result.Right - z.cx;
   taCenter :
         begin
         Result.Left  := (Result.Left + Result.Right - z.cx) div 2;
         Result.Right := (Result.Left + Result.Right + z.cx) div 2;
         end;
      end;
      case Layout of
   tlTop :
         Result.Bottom := Result.Top + z.cy;
   tlBottom  :
         Result.Top := Result.Bottom - z.cy;
   tlCenter :
         begin
         Result.Top    := (Result.Top + Result.Bottom - z.cy) div 2;
         Result.Bottom := (Result.Top + Result.Bottom + z.cy) div 2;
         end;
      end;

end;

{------------------------------------------------------------------------------}

procedure TOH_SymbLabel.SymbolTextDraw (const AText : string; const ARect : TRect; AAlign : TAlignment);

const Flags = DT_EXPANDTABS or DT_NOPREFIX;

var a, t : string;
    c, d, r : TRect;
    h : Integer;
    v : TTextLayout;
begin

   t := AText;
   r := ARect;
   repeat
      a := Uti_SubStr('$', t);
      if a <> '' then begin
         SwitchFont(a, v);
         c := r;
         h := DrawText(Canvas.Handle, PChar(a), Length(a), c, Flags or DT_CALCRECT);
         d := r;
         if v = tlBottom then d.Top := d.Bottom - h;
         DrawText(Canvas.Handle, PChar(a), Length(a), d, Flags);
         r.Left := c.Right;
         end;
      until t = '';

end;

{------------------------------------------------------------------------------}

procedure TOH_SymbLabel.DrawLabelText (var ARect : TRect; AAlign : TAlignment; const AText : string);

var t : string;
begin

   t := AText;

   Canvas.Font := Font;

   ARect := SymbolTextRect(AText, ARect, AAlign);

   if not Enabled then begin
      OffsetRect(ARect, 1, 1);
      Canvas.Font.Color := clBtnHighlight;
      SymbolTextDraw(AText, ARect, AAlign);
      OffsetRect(ARect, -1, -1);
      Canvas.Font.Color := clBtnShadow;
      SymbolTextDraw(AText, ARect, AAlign);
      end
   else
      SymbolTextDraw(AText, ARect, AAlign);

end;

{------------------------------------------------------------------------------}

procedure TOH_SymbLabel.Paint;

var Rect : TRect;
begin

   with Canvas do begin

   if not Transparent then begin
      Brush.Color := Self.Color;
      Brush.Style := bsSolid;
      FillRect(ClientRect);
      end;

   Brush.Style := bsClear;
   Rect := ClientRect;

   DrawLabelText(Rect, Alignment, Caption);

end end;

{------------------------------------------------------------------------------}

procedure TOH_SymbLabel.Changed (Sender : TObject);
begin
   Invalidate;
end;

{------------------------------------------------------------------------------}

procedure TOH_SymbLabel.SetSmallFont (Value : TFont);
begin
   FSmallFont.Assign(Value);
end;

{------------------------------------------------------------------------------}

procedure TOH_SymbLabel.SetStrongFont (Value : TFont);
begin
   FStrongFont.Assign(Value);
end;

{------------------------------------------------------------------------------}

procedure TOH_SymbLabel.SetSymbolFont (Value : TFont);
begin
   FSymbolFont.Assign(Value);
end;


{==============================================================================}

constructor TOHReadoutLabel.Create (AOwner : TComponent);
begin

   inherited Create(AOwner);

   AutoSize    := False;
//   Layout      := tlCenter;
   Color       := clWindow;
   Transparent := False;

   Width       := 120;
   FValueWidth := 40;
   FUnitsWidth := 30;

end;

{------------------------------------------------------------------------------}

procedure TOHReadoutLabel.Paint;

var ra, rc : TRect;
    c1, c2 : TColor;
begin

   with Canvas do begin

   Brush.Color := Self.Color;
   Brush.Style := bsSolid;

   rc := ClientRect;
   FillRect(rc);

   ra := rc;
   Dec(ra.Right, 8 + FValueWidth + FUnitsWidth);
//   FillRect(ra);
   DrawLabelText(ra, taRightJustify, Caption);

   ra := rc;
   Dec(ra.Right, FUnitsWidth);
   ra.Left := ra.Right - FValueWidth;
   Gfx_BevelColours(True, Color, c1, c2);
   Frame3D(Canvas, ra, c1, c2, 1);
//   FillRect(ra);
   Inc(ra.Left, 4);
   Dec(ra.Right, 4);
   DrawLabelText(ra, taRightJustify, FValueText);

   ra := rc;
   ra.Left := ra.Right - FUnitsWidth + 8;
//   FillRect(ra);
   DrawLabelText(ra, taLeftJustify, FUnitsText);

end end;

{------------------------------------------------------------------------------}

procedure TOHReadoutLabel.SetValueText (const Value : string);

var r : TRect;
begin
   FValueText := Value;

// reduce flashing by selective invalidate
   if (Visible or (csDesigning in ComponentState) and not (csNoDesignVisible in ControlStyle)) and
      (Parent <> nil) and Parent.HandleAllocated then begin
      r := BoundsRect;
      Dec(r.Right, FUnitsWidth);
      r.Left := r.Right - FValueWidth;
      InvalidateRect(Parent.Handle, @r, False);
      end;
//   Invalidate;
end;

{------------------------------------------------------------------------------}

procedure TOHReadoutLabel.SetValueWidth (Value : Integer);
begin
   if Value <> FValueWidth then begin
      FValueWidth := Value;
      Invalidate;
      end;
end;

{------------------------------------------------------------------------------}

procedure TOHReadoutLabel.SetUnitsText (const Value : string);
begin
   FUnitsText := Value;
   Invalidate;
end;

{------------------------------------------------------------------------------}

procedure TOHReadoutLabel.SetUnitsWidth (Value : Integer);
begin
   if Value <> FUnitsWidth then begin
      FUnitsWidth := Value;
      Invalidate;
      end;
end;


{==============================================================================}
{     List Label
{==============================================================================}


constructor TOHListLabel.Create (AOwner : TComponent);
begin

   inherited Create(AOwner);

   AutoSize      := False;
   ShowAccelChar := False;

//   inherited DefaultRowHeight := 20;
   FBevelWidth   := 1;
   FColumns      := CreateColumns;

//   SizeChanged(ColCount, RowCount);

   FItems := TStringList.Create;
   FItems.OnChange := ItemsChanged;

end;

{------------------------------------------------------------------------------}

destructor TOHListLabel.Destroy;
begin

   FColumns.Free;
   FColumns := nil;

   FItems.Free;

   inherited Destroy;

end;

{------------------------------------------------------------------------------}

function TOHListLabel.CreateColumns : TOHListColumns;
begin
   Result := TOHListColumns.Create(Self, TOHListColumn);
end;

{------------------------------------------------------------------------------}
{
procedure TOHListLabel.Loaded;
begin

   inherited Loaded;
   AdjustBounds;

end;

{------------------------------------------------------------------------------}

function TOHListLabel.WidthToColumn (ACol : Integer) : Integer;

var i : Integer;
begin
   Result := 0;
   for i := 0 to ACol - 1 do if i < FColumns.Count then
      Inc(Result, FColumns[i].Width);
end;

{------------------------------------------------------------------------------}

function TOHListLabel.HeightToRow (ARow : Integer) : Integer;

var h : Integer;
begin
   h := FItemHeight;  if h <= 0 then h := DefaultItemHeight;
   Result := h * ARow;
   if Result < 0 then Result := 0;
end;

{------------------------------------------------------------------------------}

procedure TOHListLabel.Paint;

const ah : array [TAlignment]  of Word = (DT_LEFT, DT_RIGHT, DT_CENTER);
      av : array [TTextLayout] of Word = (DT_TOP, DT_VCENTER, DT_BOTTOM);
      ww : array [Boolean]     of Word = (DT_SINGLELINE, DT_WORDBREAK);

var i, j, n, x1, x2, y1, y2 : Integer;
    a, t : string;
    al : TAlignment;
    c1, c2 : TColor;
    cr, rr : TRect;
begin

   with Canvas do begin

   cr := ClientRect;
   if FBorderStyle <> bsNone then begin
      Canvas.Brush.Color := clBlack;
      Canvas.FrameRect(cr);
      InflateRect(cr, -1, -1);
      end;

   if not Transparent then begin
      Brush.Color := Self.Color;
      Brush.Style := bsSolid;
      FillRect(cr);
      end;
{
   Canvas.Pen.Color := clBlack;
   x1 := cr.Left;
   for j := 0 to ColCount - 1 do with FColumns[j] do if Lined then begin
      Inc(x1, Width);
      MoveTo(x1, cr.Top);
      LineTo(x1, cr.Bottom);
      end;
}
   if FBevel <> bvNone then begin
      Gfx_BevelColours((FBevel = bvLowered), Color, c1, c2);
      rr := Rect(cr.Left + WidthToColumn(FBorderCols),
                 cr.Top + HeightToRow(FBorderRows), cr.Right, cr.Bottom);
      Frame3D(Canvas, rr, c1, c2, FBevelWidth);
      end;

   Brush.Style := bsClear;
   n := ColCount;  if n < 1 then n := 1;
   y1 := cr.Top;
   for i := 0 to FItems.Count - 1 do begin
      if (i = FBorderRows) and (FBevel <> bvNone) then
         Inc(y1, FBevelWidth);
      if FItemHeight < 1 then y2 := y1 + DefaultItemHeight
                         else y2 := y1 + FItemHeight;
      x1 := cr.Left;
      t := FItems.Strings[i];
      for j := 1 to n do begin
         if (j = FBorderCols) and (FBevel <> bvNone) then
            Inc(x1, FBevelWidth);
         if j < n then a := Uti_SubStr('|', t)
                  else a := t;
         if j > ColCount then begin
            x2 := cr.Right;
            al := Alignment;
            Canvas.Font := Font;
            end
         else with FColumns[j - 1] do begin
            x2 := x1 + Width;
            al := Alignment;
            Canvas.Font := GetDisplayFont;
            end;
         if not Enabled then Canvas.Font.Color := clGrayText;
         rr := Rect(x1 + 2, y1, x2 - 3, y2);
         DrawText(Canvas.Handle, PChar(a), Length(a), rr,
            DT_EXPANDTABS or ah[al] or av[Layout] or ww[WordWrap]);
         x1 := x2;
         end;
      y1 := y2;
      end;

end end;

{------------------------------------------------------------------------------}

procedure TOHListLabel.AdjustBounds;

var b, h, i, w : Integer;
begin

   if csReading in ComponentState then Exit;
   if not AutoSize then Exit;
   if FItems.Count = 0 then Exit;

   w := 0;
   for i := 0 to FColumns.Count - 1 do
      Inc(w, FColumns[i].Width);
   if w <= 0 then Exit;

   h := FItemHeight;  if h <= 0 then h := DefaultItemHeight;

   b := FBevelWidth + FBevelWidth;
   if FBorderStyle <> bsNone then Inc(b, 2);

   SetBounds(Left, Top, b + w, b + h * FItems.Count);

end;

{------------------------------------------------------------------------------}

procedure TOHListLabel.Clear;
begin
   FItems.Clear;
end;

{------------------------------------------------------------------------------}

procedure TOHListLabel.SetBevel (Value : TPanelBevel);
begin

   if Value <> FBevel then begin
      FBevel := Value;
      Invalidate;
      end;

end;

{------------------------------------------------------------------------------}

procedure TOHListLabel.SetBevelWidth (Value : Integer);
begin

   if Value <> FBevelWidth then begin
      FBevelWidth := Value;
      Invalidate;
      end;

end;

{------------------------------------------------------------------------------}

procedure TOHListLabel.SetBorderCols (Value : Integer);
begin

   if Value <> FBorderCols then begin
      FBorderCols := Value;
      Invalidate;
      end;

end;

{------------------------------------------------------------------------------}

procedure TOHListLabel.SetBorderRows (Value : Integer);
begin

   if Value <> FBorderRows then begin
      FBorderRows := Value;
      Invalidate;
      end;

end;

{------------------------------------------------------------------------------}

procedure TOHListLabel.SetBorderStyle (Value : TBorderStyle);
begin

   if Value <> FBorderStyle then begin
      FBorderStyle := Value;
      Invalidate;
      end;

end;

{------------------------------------------------------------------------------}
{
procedure TOHListLabel.SetBorderWidth (Value : Integer);
begin

   if Value <> FBorderWidth then begin
      FBorderWidth := Value;
      Invalidate;
      end;

end;

{------------------------------------------------------------------------------}

function TOHListLabel.GetColCount : Integer;
begin
   Result := FColumns.Count;
end;

{------------------------------------------------------------------------------}

procedure TOHListLabel.SetColCount (Value : Integer);
begin

   if csLoading in ComponentState then Exit;

   while FColumns.Count > Value do FColumns[Value].Destroy;
   while FColumns.Count < Value do FColumns.Add;

end;

{------------------------------------------------------------------------------}

procedure TOHListLabel.SetColumns (Value : TOHListColumns);
begin
   FColumns.Assign(Value);
end;

{------------------------------------------------------------------------------}

procedure TOHListLabel.SetItemHeight (Value : Integer);
begin

   if Value <> FItemHeight then begin
      FItemHeight := Value;
      Invalidate;
      end;

end;

{------------------------------------------------------------------------------}

procedure TOHListLabel.ItemsChanged (Sender : TObject);
begin
   AdjustBounds;
   Invalidate;
end;

{------------------------------------------------------------------------------}

procedure TOHListLabel.SetItems (Value : TStringList);
begin
   Items.Assign(Value);
end;

{------------------------------------------------------------------------------}

function TOHListLabel.GetRowCount : Integer;
begin
   Result := FItems.Count;
end;


{==============================================================================}
{     Column Details
{==============================================================================}

constructor TOHListColumn.Create (Collection : TCollection);

var l : TOHListLabel;
begin

   l := nil;
   if Assigned(Collection) and (Collection is TOHListColumns) then
      l := TOHListColumns(Collection).List;

   try
      inherited Create(Collection);

      FAlignment     := taLeftJustify;
      FColor         := clNone;
      FFont          := TFont.Create;
      FFont.Assign(DefaultFont);
      FFont.OnChange := FontChanged;
      FParentColor   := True;
      FParentFont    := True;
      FWidth         := 50;

   finally
      if l <> nil then begin
         l.SetColCount(List.Columns.Count);
         end;
   end;

end;

{------------------------------------------------------------------------------}

destructor TOHListColumn.Destroy;
begin

   FFont.Free;

   inherited Destroy;

end;

{------------------------------------------------------------------------------}

procedure TOHListColumn.Assign (Source : TPersistent);
begin

   if Source is TOHListColumn then begin
      if Assigned(Collection) then Collection.BeginUpdate;
      try
         RestoreDefaults;
         Color        := TOHListColumn(Source).Color;
         Font         := TOHListColumn(Source).Font;
         Alignment    := TOHListColumn(Source).Alignment;
         ParentColor  := TOHListColumn(Source).ParentColor;
         ParentFont   := TOHListColumn(Source).ParentFont;
         FWidth       := TOHListColumn(Source).FWidth;
      finally
         if Assigned(Collection) then Collection.EndUpdate;
      end;
      end
   else
      inherited Assign(Source);

end;

{------------------------------------------------------------------------------}

function TOHListColumn.DefaultColor : TColor;

var l : TOHListLabel;
begin

   l := GetList;
   if Assigned(l) then Result := l.Color
                  else Result := clWindow;

end;

{------------------------------------------------------------------------------}

function TOHListColumn.DefaultFont : TFont;

var l : TOHListLabel;
begin

   l := GetList;
   if Assigned(l) then Result := l.Font
                  else Result := FFont;

end;

{------------------------------------------------------------------------------}

function TOHListColumn.IsColorStored : Boolean;
begin
   Result := not FParentColor;
end;

{------------------------------------------------------------------------------}

function TOHListColumn.GetDisplayFont : TFont;

var l : TOHListLabel;
begin

   l := GetList;
   if not Assigned(l) or {(csDesigning in l.ComponentState) or} not FParentFont then
      Result := FFont
   else
      Result := l.Font;

end;

{------------------------------------------------------------------------------}

function TOHListColumn.GetList : TOHListLabel;
begin
   if Assigned(Collection) and (Collection is TOHListColumns) then
      Result := TOHListColumns(Collection).List
   else
      Result := nil;
end;

{------------------------------------------------------------------------------}

function TOHListColumn.IsFontStored : Boolean;
begin
   Result := not FParentFont;
end;

{------------------------------------------------------------------------------}

procedure TOHListColumn.RestoreDefaults;
begin
   FParentColor := True;
   FParentFont  := True;
   Changed(False);
end;

{------------------------------------------------------------------------------}

procedure TOHListColumn.SetAlignment (Value : TAlignment);
begin

   if Value <> FAlignment then begin
      FAlignment := Value;
      Changed(False);
      end;

end;

{------------------------------------------------------------------------------}

procedure TOHListColumn.SetColor (Value : TColor);
begin

   if (Value <> FColor) or FParentColor then begin
      FColor := Value;
      FParentColor := False;
      Changed(False);
      end;

end;

{------------------------------------------------------------------------------}

procedure TOHListColumn.SetFont (Value : TFont);
begin

   FFont.Assign(Value);
   Changed(False);

end;

{------------------------------------------------------------------------------}

procedure TOHListColumn.FontChanged;
begin
   FParentFont := False;
   Changed(False);
end;

{------------------------------------------------------------------------------}

procedure TOHListColumn.SetLined (Value : Boolean);
begin

   if Value <> FLined then begin
      FLined := Value;
      Changed(False);
      end;

end;

{------------------------------------------------------------------------------}

procedure TOHListColumn.SetParentColor (Value : Boolean);
begin

   if Value <> FParentColor then begin
      FParentColor := Value;
      Changed(False);
      end;

end;

{------------------------------------------------------------------------------}

procedure TOHListColumn.SetParentFont (Value : Boolean);
begin

   if Value <> FParentFont then begin
      FParentFont := Value;
      Changed(False);
      end;

end;

{------------------------------------------------------------------------------}

procedure TOHListColumn.SetWidth (Value : Integer);
begin

   if Value <> FWidth then begin
      FWidth := Value;
      Changed(False);
      end;

end;


{==============================================================================}
{
{==============================================================================}

constructor TOHListColumns.Create (List : TOHListLabel; ColumnClass : TOHListColumnClass);
begin
   inherited Create(ColumnClass);
   FList := List;
end;

{------------------------------------------------------------------------------}

function TOHListColumns.Add : TOHListColumn;
begin
   Result := TOHListColumn(inherited Add);
end;

{------------------------------------------------------------------------------}

function TOHListColumns.GetColumn (Index : Integer) : TOHListColumn;
begin
   Result := TOHListColumn(inherited Items[Index]);
end;

{------------------------------------------------------------------------------}

function TOHListColumns.GetOwner : TPersistent;
begin
   Result := FList;
end;

{------------------------------------------------------------------------------}

procedure TOHListColumns.LoadFromFile(const Filename : string);

var s : TFileStream;
begin

   s := TFileStream.Create(Filename, fmOpenRead);
   try
      LoadFromStream(s);
   finally
      s.Free;
   end;

end;

{------------------------------------------------------------------------------}

type TColumnsWrapper = class(TComponent)
   private
      FColumns : TOHListColumns;
   published
      property Columns : TOHListColumns read FColumns write FColumns;
   end;

{------------------------------------------------------------------------------}

procedure TOHListColumns.LoadFromStream (S : TStream);

var w : TColumnsWrapper;
begin

   w := TColumnsWrapper.Create(nil);
   try
      w.Columns := FList.CreateColumns;
      S.ReadComponent(w);
      Assign(w.Columns);
   finally
      w.Columns.Free;
      w.Free;
   end;

end;

{------------------------------------------------------------------------------}

procedure TOHListColumns.RestoreDefaults;

var i : Integer;
begin

   BeginUpdate;
   try
      for i := 0 to Count - 1 do Items[i].RestoreDefaults;
   finally
      EndUpdate;
   end;

end;

{------------------------------------------------------------------------------}

procedure TOHListColumns.SaveToFile (const Filename : string);

var s : TStream;
begin

   s := TFileStream.Create(Filename, fmCreate);
   try
      SaveToStream(s);
   finally
      s.Free;
   end;

end;

{------------------------------------------------------------------------------}

procedure TOHListColumns.SaveToStream (S : TStream);

var w : TColumnsWrapper;
begin

   w := TColumnsWrapper.Create(nil);
   try
      w.Columns := Self;
      S.WriteComponent(w);
   finally
      w.Free;
   end;

end;

{------------------------------------------------------------------------------}

procedure TOHListColumns.SetColumn (Index : Integer; Value : TOHListColumn);
begin
   Items[Index].Assign(Value);
end;

{------------------------------------------------------------------------------}

procedure TOHListColumns.Update (Item : TCollectionItem);
begin

   if (FList <> nil) and not (csLoading in FList.ComponentState) then
      FList.Invalidate;

end;


{==============================================================================}

procedure Register;
begin
   RegisterComponents('Orchard House', [TOHURLLabel, TOHHotLabel]);
   RegisterComponents('Orchard House', [TOHSymbLabel, TOHReadoutLabel]);
   RegisterComponents('Orchard House', [TOHListLabel]);
end;

{==============================================================================}

end.
